@@ -2,22 +2,39 @@ from __future__ import print_function
 from logging import getLogger
 logger = getLogger(__name__)
 
-import thread
+try:
+    import thread
+except ImportError: # python 3
+    import _thread as thread
 import threading
 import traceback
 import socket
 import struct
-import select
 import sys
 import os
 import atexit
 import code
 import signal
+import errno
 try:
     import signalfd
 except ImportError:
     signalfd = None
-
+try:
+    string = basestring
+except NameError: # python 3
+    string = str
+try:
+    InterruptedError = InterruptedError
+except NameError: # python <= 3.2
+    InterruptedError = OSError
+if hasattr(sys, 'setswitchinterval'):
+    setinterval = sys.setswitchinterval
+    getinterval = sys.getswitchinterval
+else:
+    setinterval = sys.setcheckinterval
+    getinterval = sys.getcheckinterval
+PY3 = sys.version_info[0] == 3
 VERBOSE = True
 
 try:
@@ -32,7 +49,7 @@ try:
     _pthread_setname_np = libpthread.pthread_setname_np
     _pthread_setname_np.argtypes = [ctypes.c_void_p, ctypes.c_char_p]
     _pthread_setname_np.restype = ctypes.c_int
-    pthread_setname_np = lambda ident, name: _pthread_setname_np(ident, name[:15])
+    pthread_setname_np = lambda ident, name: _pthread_setname_np(ident, name[:15].encode('utf8'))
 except ImportError:
     pthread_setname_np = lambda ident, name: None
 
@@ -87,23 +104,19 @@ class Manhole(threading.Thread):
         pthread_setname_np(self.ident, self.name)
 
         sock, pid = self.get_socket()
-        cry("Waiting for new connection (in pid:%s) ..." % pid)
         while True:
-            client, _ = sock.accept()
-            global _CLIENT_INST #pylint: disable=W0603
-
+            cry("Waiting for new connection (in pid:%s) ..." % pid)
             try:
-                _CLIENT_INST = ManholeConnection(client, self.sigmask)
-                _CLIENT_INST.start()
-                _CLIENT_INST.join()
-            #except: #pylint: disable=W0703
-            #    cry(traceback.format_exc()) #pylint: disable=W0702
+                client = ManholeConnection(sock.accept()[0], self.sigmask)
+                client.start()
+                client.join()
+            except (InterruptedError, socket.error) as e:
+                if e.errno != errno.EINTR:
+                    raise
+                continue
             finally:
-                _CLIENT_INST = None
                 del client
 
-            cry("Waiting for new connection ...")
-
 class ManholeConnection(threading.Thread):
     def __init__(self, client, sigmask):
         super(ManholeConnection, self).__init__()
@@ -117,12 +130,9 @@ class ManholeConnection(threading.Thread):
             signalfd.sigprocmask(signalfd.SIG_BLOCK, self.sigmask)
         pthread_setname_np(self.ident, "Manhole ----")
 
-        client = self.client
-        del self.client
-        pid, _, _ = self.check_credentials(client)
+        pid, _, _ = self.check_credentials(self.client)
         pthread_setname_np(self.ident, "Manhole %s" % pid)
-
-        self.handle(client)
+        self.handle(self.client)
 
     @staticmethod
     def check_credentials(client):
@@ -145,43 +155,51 @@ class ManholeConnection(threading.Thread):
         client.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 0)
         client.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 0)
         backup = []
+        old_interval = getinterval()
         try:
-            client_fd = client.fileno()
-            for mode, names in (
-                ('w', (
-                    'stderr',
-                    'stdout',
-                    '__stderr__',
-                    '__stdout__'
-                )),
-                ('r', (
-                    'stdin',
-                    '__stdin__'
-                ))
-            ):
-                for name in names:
-                    backup.append((name, getattr(sys, name)))
-                    setattr(sys, name, os.fdopen(client_fd, mode, 0))
-
-            run_repl()
-            cry("DONE.")
-        finally:
-            cry("Cleaning up.")
-            old_interval = sys.getcheckinterval()
-            sys.setcheckinterval(2147483647)
-            junk = [] # keep the old file objects alive for a bit
-            for name, fh in backup:
-                junk.append(getattr(sys, name))
-                setattr(sys, name, fh)
-            del backup
-            for fh in junk:
+            try:
+                client_fd = client.fileno()
+                for mode, names in (
+                    ('w', (
+                        'stderr',
+                        'stdout',
+                        '__stderr__',
+                        '__stdout__'
+                    )),
+                    ('r', (
+                        'stdin',
+                        '__stdin__'
+                    ))
+                ):
+                    for name in names:
+                        backup.append((name, getattr(sys, name)))
+                        setattr(sys, name, os.fdopen(client_fd, mode, 1 if PY3 else 0))
+                run_repl()
+                cry("DONE.")
+            finally:
                 try:
-                    fh.close()
-                except IOError:
-                    pass
-                del fh
-            del junk
-            sys.setcheckinterval(old_interval)
+                    setinterval(2147483647) # change the switch/check interval to something ridiculous
+                                            # we don't want to have other thread try to write to the
+                                            # redirected sys.__std*/sys.std* - it would fail horribly
+                    client.close() # close before it's too late. it may already be dead
+                    junk = [] # keep the old file objects alive for a bit
+                    for name, fh in backup:
+                        junk.append(getattr(sys, name))
+                        setattr(sys, name, fh)
+                    del backup
+                    for fh in junk:
+                        try:
+                            fh.close()
+                        except IOError:
+                            pass
+                        del fh
+                    del junk
+                finally:
+                    setinterval(old_interval)
+                    cry("Cleaned up.")
+        except Exception:
+            cry("ManholeConnection thread failed:")
+            cry(traceback.format_exc())
 
 def run_repl():
     dump_stacktraces()
@@ -200,8 +218,10 @@ def _handle_oneshot(_signum, _frame):
         client, _ = sock.accept()
         ManholeConnection.check_credentials(client)
         ManholeConnection.handle(client)
-    except SystemExit:
-        pass
+    except: # pylint: disable=W0702
+        # we don't want to let any exception out, it might make the application missbehave
+        cry("Manhole oneshot connection failed:")
+        cry(traceback.format_exc())
 
 def _remove_manhole_uds():
     name = "/tmp/manhole-%s" % os.getpid()
@@ -209,7 +229,7 @@ def _remove_manhole_uds():
         os.unlink(name)
 
 _INST_LOCK = thread.allocate_lock()
-_STDERR = _INST = _CLIENT_INST = _ORIGINAL_OS_FORK = _ORIGINAL_OS_FORKPTY = _SHOULD_RESTART = None
+_STDERR = _INST = _ORIGINAL_OS_FORK = _ORIGINAL_OS_FORKPTY = _SHOULD_RESTART = None
 
 def _patched_fork():
     """Fork a child process."""
@@ -236,7 +256,7 @@ def _patch_os_fork_functions():
     cry("Patched %s and %s." % (_ORIGINAL_OS_FORK, _ORIGINAL_OS_FORKPTY))
 
 def _activate_on_signal(_signum, _frame):
-    assert _INST
+    assert _INST, "Manhole wasn't installed !"
     _INST.start()
 
 ALL_SIGNALS = [
@@ -251,7 +271,7 @@ def install(verbose=True, patch_fork=True, activate_on=None, sigmask=ALL_SIGNALS
         if not _INST:
             _INST = Manhole(sigmask)
             if oneshot_on is not None:
-                oneshot_on = getattr(signal, 'SIG'+oneshot_on) if isinstance(oneshot_on, basestring) else oneshot_on
+                oneshot_on = getattr(signal, 'SIG'+oneshot_on) if isinstance(oneshot_on, string) else oneshot_on
                 signal.signal(oneshot_on, _handle_oneshot)
 
             if activate_on is None:
@@ -259,7 +279,7 @@ def install(verbose=True, patch_fork=True, activate_on=None, sigmask=ALL_SIGNALS
                     _INST.start()
                     _SHOULD_RESTART = True
             else:
-                activate_on = getattr(signal, 'SIG'+activate_on) if isinstance(activate_on, basestring) else activate_on
+                activate_on = getattr(signal, 'SIG'+activate_on) if isinstance(activate_on, string) else activate_on
                 if activate_on == oneshot_on:
                     raise RuntimeError('You cannot do activation of the Manhole thread on the same signal that you want to do oneshot activation !')
                 signal.signal(activate_on, _activate_on_signal)
@@ -300,9 +320,12 @@ def dump_stacktraces():
 if __name__ == '__main__': #pragma: no cover
     from logging import basicConfig, DEBUG
     basicConfig(level=DEBUG)
-    install(verbose=True, oneshot_on='USR2')
-    import faulthandler
-    faulthandler.enable()
+    install(verbose=True)#, oneshot_on='USR2')
+    try:
+        import faulthandler
+        faulthandler.enable()
+    except ImportError:
+        pass
 
     print()
 
