@@ -1,3 +1,4 @@
+from __future__ import print_function
 import unittest
 import os
 import select
@@ -12,8 +13,12 @@ import logging
 import re
 import atexit
 import signal
-from contextlib import contextmanager
-from cStringIO import StringIO
+from contextlib import contextmanager, closing
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from io import StringIO
+
 
 class BufferingBase(object):
 
@@ -31,8 +36,8 @@ class BufferingBase(object):
                 data = os.read(self.fd, self.BUFFSIZE)
                 if not data:
                     break
-                self.buff.write(data)
-        except OSError, e:
+                self.buff.write(data.decode('utf8'))
+        except OSError as e:
             if e.errno not in (
                 errno.EAGAIN, errno.EWOULDBLOCK,
                 errno.EINPROGRESS
@@ -50,6 +55,7 @@ class TestProcess(BufferingBase):
             stdout=subprocess.PIPE,
             stderr=subprocess.STDOUT,
             env=os.environ,
+            bufsize=1,
         )
         fd = self.proc.stdout.fileno()
         flags = fcntl.fcntl(fd, fcntl.F_GETFL)
@@ -77,11 +83,20 @@ class TestProcess(BufferingBase):
                 time.sleep(0.1)
                 if self.proc.poll() is not None:
                     return
-            print >> sys.stderr, 'KILLED %s' % self
+            print('KILLED %s' % self, file=sys.stderr)
             self.proc.kill()
         except OSError as exc:
             if exc.errno != errno.ESRCH:
                 raise
+        finally:
+            self.read()
+            if self.proc.stdout:
+                self.proc.stdout.close()
+            if self.proc.stderr:
+                self.proc.stderr.close()
+            if self.proc.stdin:
+                self.proc.stdin.close()
+            self.proc.wait() # reap the zombie
 
 class TestSocket(BufferingBase):
     BUFFSIZE = 8192
@@ -135,9 +150,9 @@ class ManholeTestCase(unittest.TestCase):
         try:
             yield
         except Exception:
-            print "*********** OUTPUT ***********"
-            print cb()
-            print "******************************"
+            print("*********** OUTPUT ***********")
+            print(cb())
+            print("******************************")
             raise
 
     def test_simple_r01(self):
@@ -164,57 +179,59 @@ class ManholeTestCase(unittest.TestCase):
     def run_simple(self, count):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_simple') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 2, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 for _ in range(count):
                     proc.reset()
                     self.assertManholeRunning(proc, uds_path)
 
     def assertManholeRunning(self, proc, uds_path, oneshot=False):
         sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-        sock.settimeout(0.05)
+        sock.settimeout(0.5)
         sock.connect(uds_path)
-        with TestSocket(sock) as client:
-            with self._dump_on_error(client.read):
-                self._wait_for_strings(client.read, 1,
-                    "ProcessID",
-                    "ThreadID",
-                    ">>>",
-                )
-                sock.send("print 'FOOBAR'\n")
-                self._wait_for_strings(client.read, 1, "FOOBAR")
+        try:
+            with TestSocket(sock) as client:
+                with self._dump_on_error(client.read):
+                    self._wait_for_strings(client.read, 10,
+                        "ProcessID",
+                        "ThreadID",
+                        ">>>",
+                    )
+                    sock.send(b"print('FOOBAR')\n")
+                    self._wait_for_strings(client.read, 10, "FOOBAR")
 
-                self._wait_for_strings(proc.read, 1,
-                    'from PID:%s UID:%s' % (os.getpid(), os.getuid()),
-                )
-                sock.shutdown(socket.SHUT_RDWR)
-                sock.close()
-        self._wait_for_strings(proc.read, 1,
-            'Cleaning up.',
+                    self._wait_for_strings(proc.read, 10,
+                        'from PID:%s UID:%s' % (os.getpid(), os.getuid()),
+                    )
+                    sock.shutdown(socket.SHUT_RDWR)
+        finally:
+            sock.close()
+        self._wait_for_strings(proc.read, 10,
+            'Cleaned up.',
             *[] if oneshot else ['Waiting for new connection']
         )
     def test_exit_with_grace(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_simple') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
 
                 sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                 sock.settimeout(0.05)
                 sock.connect(uds_path)
                 with TestSocket(sock) as client:
                     with self._dump_on_error(client.read):
-                        self._wait_for_strings(client.read, 1,
+                        self._wait_for_strings(client.read, 10,
                             "ThreadID",
                             "ProcessID",
                             ">>>",
                         )
-                        sock.send("print 'FOOBAR'\n")
-                        self._wait_for_strings(client.read, 1, "FOOBAR")
+                        sock.send(b"print('FOOBAR')\n")
+                        self._wait_for_strings(client.read, 10, "FOOBAR")
 
-                        self._wait_for_strings(proc.read, 1,
+                        self._wait_for_strings(proc.read, 10,
                             'from PID:%s UID:%s' % (os.getpid(), os.getuid()),
                         )
                         sock.shutdown(socket.SHUT_WR)
@@ -222,69 +239,72 @@ class ManholeTestCase(unittest.TestCase):
                         sock.recv(1024)
                         sock.shutdown(socket.SHUT_RD)
                         sock.close()
-                self._wait_for_strings(proc.read, 1,
+                self._wait_for_strings(proc.read, 10,
                     'DONE.',
-                    'Cleaning up.',
+                    'Cleaned up.',
                     'Waiting for new connection'
                 )
 
     def test_with_fork(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_with_fork') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 for _ in range(2):
                     proc.reset()
                     self.assertManholeRunning(proc, uds_path)
 
                 proc.reset()
-                self._wait_for_strings(proc.read, 3, 'Fork detected')
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, 'Fork detected')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 new_uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self.failIfEqual(uds_path, new_uds_path)
+                self.assertNotEqual(uds_path, new_uds_path)
 
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 for _ in range(2):
                     proc.reset()
                     self.assertManholeRunning(proc, new_uds_path)
+    if not hasattr(sys, 'pypy_version_info'):
+        def test_with_forkpty(self):
+            with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_with_forkpty') as proc:
+                with self._dump_on_error(proc.read):
+                    self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
+                    uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
+                    self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                    for _ in range(2):
+                        proc.reset()
+                        self.assertManholeRunning(proc, uds_path)
 
-    def test_with_forkpty(self):
-        with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_with_forkpty') as proc:
-            with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
-                uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
-                for _ in range(2):
                     proc.reset()
-                    self.assertManholeRunning(proc, uds_path)
+                    self._wait_for_strings(proc.read, 10, 'Fork detected')
+                    self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
+                    new_uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
+                    self.assertNotEqual(uds_path, new_uds_path)
 
-                proc.reset()
-                self._wait_for_strings(proc.read, 3, 'Fork detected')
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
-                new_uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self.failIfEqual(uds_path, new_uds_path)
-
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
-                for _ in range(2):
-                    proc.reset()
-                    self.assertManholeRunning(proc, new_uds_path)
+                    self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                    for _ in range(2):
+                        proc.reset()
+                        self.assertManholeRunning(proc, new_uds_path)
 
     def test_auth_fail(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_auth_fail') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
-                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
-                sock.settimeout(1)
-                sock.connect(uds_path)
-                self.assertEquals("", sock.recv(1024))
-                self._wait_for_strings(proc.read, 1,
-                    "SuspiciousClient: Can't accept client with PID:-1 UID:-1 GID:-1. It doesn't match the current EUID:",
-                    'Waiting for new connection'
-                )
-                proc.proc.send_signal(signal.SIGINT)
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                with closing(socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)) as sock:
+                    sock.settimeout(1)
+                    sock.connect(uds_path)
+                    try:
+                        self.assertEqual(b"", sock.recv(1024))
+                    except socket.timeout:
+                        pass
+                    self._wait_for_strings(proc.read, 10,
+                        "SuspiciousClient: Can't accept client with PID:-1 UID:-1 GID:-1. It doesn't match the current EUID:",
+                        'Waiting for new connection'
+                    )
+                    proc.proc.send_signal(signal.SIGINT)
 
     try:
         import signalfd
@@ -294,49 +314,49 @@ class ManholeTestCase(unittest.TestCase):
         def test_signalfd_weirdness(self):
             with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_signalfd_weirdness') as proc:
                 with self._dump_on_error(proc.read):
-                    self._wait_for_strings(proc.read, 2, '/tmp/manhole-')
+                    self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                     uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                    self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
-                    self._wait_for_strings(proc.read, 2, *[
-                        '[%s] read from signalfd:' % j for j in range(100)
+                    self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                    self._wait_for_strings(proc.read, 250, *[
+                        '[%s] read from signalfd:' % j for j in range(200)
                     ])
                     self.assertManholeRunning(proc, uds_path)
 
         def test_signalfd_weirdness_negative(self):
             with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_signalfd_weirdness_negative') as proc:
                 with self._dump_on_error(proc.read):
-                    self._wait_for_strings(proc.read, 2, '/tmp/manhole-')
+                    self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                     uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                    self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
-                    self.assertRaises(AssertionError, self._wait_for_strings, proc.read, 2, '[0] read from signalfd:')
+                    self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
+                    self._wait_for_strings(proc.read, 10, 'reading from signalfd failed')
                     self.assertManholeRunning(proc, uds_path)
 
 
     def test_activate_on_usr2(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_activate_on_usr2') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, 'Not patching os.fork and os.forkpty. Activation is done by signal 12')
+                self._wait_for_strings(proc.read, 10, 'Not patching os.fork and os.forkpty. Activation is done by signal 12')
                 self.assertRaises(AssertionError, self._wait_for_strings, proc.read, 2, '/tmp/manhole-')
                 proc.signal(signal.SIGUSR2)
-                self._wait_for_strings(proc.read, 2, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 self.assertManholeRunning(proc, uds_path)
 
     def test_activate_on_with_oneshot_on(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_activate_on_with_oneshot_on') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, "RuntimeError('You cannot do activation of the Manhole thread on the same signal that you want to do oneshot activation !')")
+                self._wait_for_strings(proc.read, 10, "RuntimeError('You cannot do activation of the Manhole thread on the same signal that you want to do oneshot activation !')")
 
     def test_oneshot_on_usr2(self):
         with TestProcess(sys.executable, '-u', __file__, 'daemon', 'test_oneshot_on_usr2') as proc:
             with self._dump_on_error(proc.read):
-                self._wait_for_strings(proc.read, 1, 'Not patching os.fork and os.forkpty. Oneshot activation is done by signal 12')
+                self._wait_for_strings(proc.read, 10, 'Not patching os.fork and os.forkpty. Oneshot activation is done by signal 12')
                 self.assertRaises(AssertionError, self._wait_for_strings, proc.read, 2, '/tmp/manhole-')
                 proc.signal(signal.SIGUSR2)
-                self._wait_for_strings(proc.read, 2, '/tmp/manhole-')
+                self._wait_for_strings(proc.read, 10, '/tmp/manhole-')
                 uds_path = re.findall("(/tmp/manhole-\d+)", proc.read())[0]
-                self._wait_for_strings(proc.read, 1, 'Waiting for new connection')
+                self._wait_for_strings(proc.read, 10, 'Waiting for new connection')
                 self.assertManholeRunning(proc, uds_path, oneshot=True)
 
 cov = None
@@ -346,13 +366,16 @@ def maybe_enable_coverage():
         cov.stop()
     cov = cov or os.environ.get("WITH_COVERAGE")
     if cov:
-        from coverage.control import coverage
+        try:
+            from coverage.control import coverage
+        except ImportError:
+            cov = None
+            return
         cov = coverage(auto_data=True, data_suffix=True, timid=False, include=['src/*'])
         cov.start()
 
-    @atexit.register
-    def cleanup():
-        if cov:
+        @atexit.register
+        def cleanup():
             cov.stop()
             cov.save()
 
@@ -387,7 +410,8 @@ if __name__ == '__main__':
             return pid
 
         @monkeypatch(os, 'forkpty')
-        def patched_forkpty((pid, fd)):
+        def patched_forkpty(pid_fd):
+            pid, fd = pid_fd
             if not pid:
                 maybe_enable_coverage()
             return pid, fd
@@ -411,29 +435,31 @@ if __name__ == '__main__':
                 manhole.install(sigmask=None)
             else:
                 manhole.install(sigmask=[signal.SIGCHLD])
-            print 'Starting ...'
+            print('Starting ...')
             import signalfd
             signalfd.sigprocmask(signalfd.SIG_BLOCK, [signal.SIGCHLD])
             fd = signalfd.signalfd(0, [signal.SIGCHLD], signalfd.SFD_NONBLOCK|signalfd.SFD_CLOEXEC)
-            for i in range(100):
-                print 'Forking', i
+            for i in range(200):
+                print('Forking', i)
                 pid = os.fork()
-                print ' - [%s/%s] forked' % (i, pid)
+                print(' - [%s/%s] forked' % (i, pid))
                 if pid:
                     while 1:
-                        print ' - [%s/%s] selecting on: %s' % (i, pid, [fd])
+                        print(' - [%s/%s] selecting on: %s' % (i, pid, [fd]))
                         read_ready, _, errors = select.select([fd], [], [fd], 1)
                         if read_ready:
                             try:
-                                print ' - [%s/%s] reading from signalfd ...' % (i, pid)
-                                print ' - [%s] read from signalfd: %r ' % (i, os.read(fd, 128))
+                                print(' - [%s/%s] reading from signalfd ...' % (i, pid))
+                                print(' - [%s] read from signalfd: %r ' % (i, os.read(fd, 128)))
                                 break
                             except OSError as exc:
-                                print ' - [%s/%s] reading from signalfd failed with errno %s' % (i, pid, exc.errno)
+                                print(' - [%s/%s] reading from signalfd failed with errno %s' % (i, pid, exc.errno))
+                        else:
+                            print(' - [%s/%s] reading from signalfd failed - not ready !' % (i, pid))
                         if errors:
                             raise RuntimeError("fd has error")
                 else:
-                    print ' - [%s/%s] exiting' % (i, pid)
+                    print(' - [%s/%s] exiting' % (i, pid))
                     os._exit(0)
             time.sleep(1)
         else:
@@ -446,11 +472,18 @@ if __name__ == '__main__':
                 if pid:
                     @atexit.register
                     def cleanup():
-                        os.kill(pid, signal.SIGINT)
-                        time.sleep(0.2)
-                        os.kill(pid, signal.SIGTERM)
+                        try:
+                            os.kill(pid, signal.SIGINT)
+                            time.sleep(0.2)
+                            os.kill(pid, signal.SIGTERM)
+                        except OSError as e:
+                            if e.errno != errno.ESRCH:
+                                raise
                     while not os.waitpid(pid, os.WNOHANG)[0]:
-                        os.write(2, os.read(masterfd, 1024))
+                        try:
+                            os.write(2, os.read(masterfd, 1024))
+                        except OSError as e:
+                            print("Error while reading from masterfd:", e)
                 else:
                     time.sleep(10)
             elif test_name == 'test_with_fork':
@@ -459,9 +492,13 @@ if __name__ == '__main__':
                 if pid:
                     @atexit.register
                     def cleanup():
-                        os.kill(pid, signal.SIGINT)
-                        time.sleep(0.2)
-                        os.kill(pid, signal.SIGTERM)
+                        try:
+                            os.kill(pid, signal.SIGINT)
+                            time.sleep(0.2)
+                            os.kill(pid, signal.SIGTERM)
+                        except OSError as e:
+                            if e.errno != errno.ESRCH:
+                                raise
                     os.waitpid(pid, 0)
                 else:
                     time.sleep(10)
@@ -470,6 +507,6 @@ if __name__ == '__main__':
                 time.sleep(10)
             else:
                 raise RuntimeError('Invalid test spec.')
-        print 'DIED.'
+        print('DIED.')
     else:
         unittest.main()
